// Simple assembler for programmable constructor language:
//
// NORTH (N)
// EAST (E)
// SOUTH (S)
// WEST (W)
// FRONT (F)
// BACK (B)
//
// ORIENT dir (O dir)
//
// SLI
// FUS
// UFS
// NOR
// DEL
// ROT
//
// NOP
//
// GATHER
//
// CALL label
// RETURN (RET)
//
// label:
//
// END

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <map>

using namespace std;

// This map contains the addresses of labels
map<string,unsigned> labels;
map<unsigned,string> revlabels;

// This map maps labels onto all places where they are referred to
map<string,vector<unsigned> > labelrefs;

#define MEM_SIZE 8192

unsigned char memmap[MEM_SIZE];
unsigned address = 0, addressend = 0; // addressend is one more than the last address used
									  // (just as with iterators, end() is one past the last)
#if 0
#define M_N     0x00000009
#define M_S     0x0000000d
#define M_F     0x0000000c
#define M_B     0x00000008
#define M_W     0x0000000a
#define M_E     0x0000000e
#else
#define M_N     0x00000008
#define M_S     0x0000000c
#define M_F     0x0000000e
#define M_B     0x0000000a
#define M_W     0x00000009
#define M_E     0x0000000d
#endif

#define B_SLI   0x00000007
#define B_PSH   0x00000001
#define B_ROT   0x00000004
#define B_DEL   0x00000005
#define B_NOR   0x00000002
#define B_FUS   0x00000003
#define B_UFS   0x00000006

#define ORI     0x00000020

#define CALL    0x00000020

#define RETURN  0x00000018
#define MEMBANK 0x0000001c
#define GATHER  0x00000010
#define NOP     0x00000000

struct { char *s; unsigned n; } dirlookup[] =
{
		{"SF",0},
		{"SB",1},
		{"NB",2},
		{"NF",3},
		{"SW",4},
		{"SE",5},
		{"NE",6},
		{"NW",7},
		{"FN",8},
		{"FS",9},
		{"BS",10},
		{"BN",11},
		{"WN",12},
		{"WS",13},
		{"ES",14},
		{"EN",15},
		{"FW",20},
		{"FE",21},
		{"BE",22},
		{"BW",23},
		{"WB",28},
		{"WF",29},
		{"EF",30},
		{"EB",31},
		{"",0},
};

#define FLIP5(a) ((!!(a&1))*16 + (!!(a&2))*8 + (a&4) + (!!(a&8))*2 + (!!(a&16)))

unsigned GetDirection(string &s)
{

	for(unsigned i = 0; dirlookup[i].s[0]; i++)
	{
		if (s==dirlookup[i].s)
		{
			return FLIP5(~dirlookup[i].n);
		}
	}

	cerr << "Unrecognised direction string " << s << endl;

	return 0;
}

string GetDirectionString(unsigned d)
{
	for(unsigned i = 0; dirlookup[i].s[0]; i++)
	{
		if (d==FLIP5(~dirlookup[i].n))
		{
			return dirlookup[i].s;
		}
	}

	return "??";
}

void WriteCallInstruction(unsigned address, unsigned calladdr)
{
	// TODO - could check that it is a multiple of 4

	calladdr = calladdr >> 2;

	memmap[address++] = CALL + (calladdr&0x1f); 
	memmap[address++] = CALL + ((calladdr&0x000003e0)>>5);

}

void Assemble(void)
{
	bool done = false;
	string token;

	while(cin && !done)
	{
		cin >> token;

		if (token == "END")
		{
			done = true;
		}
		else if (token == "CALL")
		{
			cin >> token;

			if (labels.find(token) == labels.end())
			{
				labelrefs[token].push_back(address);
		
				memmap[address++] = CALL;
				address++;
			}
			else
			{
				unsigned calladdr = labels[token];

				WriteCallInstruction(address,calladdr);
				address += 2;
			}

		}
		else if (token == "RETURN" || token == "RET")
		{
			memmap[address++] = RETURN;
		}
		else if (token == "MEMBANK" || token == "MBK")
		{
			memmap[address++] = MEMBANK;
		}
		else if (token == "NORTH" || token == "N")
		{
			memmap[address++] = M_N;
		}
		else if (token == "EAST" || token == "E")
		{
			memmap[address++] = M_E;
		}
		else if (token == "SOUTH" || token == "S")
		{
			memmap[address++] = M_S;
		}
		else if (token == "WEST" || token == "W")
		{
			memmap[address++] = M_W;
		}
		else if (token == "FRONT" || token == "F")
		{
			memmap[address++] = M_F;
		}
		else if (token == "BACK" || token == "B")
		{
			memmap[address++] = M_B;
		}
		else if (token == "ORIENT" || token == "O")
		{
			cin >> token;

			memmap[address++] = ORI + GetDirection(token);
		}
		else if (token == "SLI")
		{
			memmap[address++] = B_SLI;
		}
		else if (token == "FUS")
		{
			memmap[address++] = B_FUS;
		}
		else if (token == "UFS")
		{
			memmap[address++] = B_UFS;
		}
		else if (token == "ROT")
		{
			memmap[address++] = B_ROT;
		}
		else if (token == "NOR")
		{
			memmap[address++] = B_NOR;
		}
		else if (token == "DEL")
		{
			memmap[address++] = B_DEL;
		}
		else if (token == "NOP")
		{
			memmap[address++] = NOP;
		}
		else if (token == "GATHER")
		{
			memmap[address++] = GATHER;
		}
		else if (token == "ORG")
		{
			cin >> address;
		}
		else if (token[token.length()-1] == ':')
		{
			while(address % 4 != 0)
				address++;

			token = token.substr(0,token.length()-1);

			labels[token] = address;

			if (labelrefs.find(token) != labelrefs.end())
			{
				for(vector<unsigned>::iterator i = labelrefs[token].begin(); i != labelrefs[token].end(); i++)
				{
					WriteCallInstruction(*i,address);
				}
			}
		}
		else if (token.size())
		{
			cerr << "Unrecognised token " << token << endl;
		}

		if (address > addressend)
			addressend = address;

	}

	if (!done)
	{
		cerr << "Assembly finished before END found" << endl;
		cerr << "Last token was '" << token << "'" << endl;
	}
}

unsigned DisAssemble(ofstream &os, unsigned address)
{
	unsigned tmp = address;
	unsigned w = 4;

	while(tmp)
	{
		tmp = tmp / 10;
		w = w - 1;
	}

	while(w)
	{
		os << "0";
		w = w - 1;
	}

	if (address)
	{
		os << address;
	}

	os << " ";

	switch (memmap[address])
	{
	case M_N:
		os << "NORTH" << endl;
		break;
	case M_S:
		os << "SOUTH" << endl;
		break;
	case M_E:
		os << "EAST" << endl;
		break;
	case M_W:
		os << "WEST" << endl;
		break;
	case M_F:
		os << "FRONT" << endl;
		break;
	case M_B:
		os << "BACK" << endl;
		break;
	case B_SLI:
		os << "SLI" << endl;
		break;
	case B_PSH:
		os << "PSH" << endl;
		break;
	case B_ROT:
		os << "ROT" << endl;
		break;
	case B_DEL:
		os << "DEL" << endl;
		break;
	case B_NOR:
		os << "NOR" << endl;
		break;
	case B_FUS:
		os << "FUS" << endl;
		break;
	case B_UFS:
		os << "UFS" << endl;
		break;
	case RETURN:
		os << "RETURN" << endl;
		break;
	case GATHER:
		os << "GATHER" << endl;
		break;
	case NOP:
		os << "NOP" << endl;
		break;
	case MEMBANK:
		os << "MEMBANK" << endl;
		break;
	}

	if (memmap[address] & 0x00000020)
	{
		if (address == addressend-1 || (memmap[address+1] & 0x00000020) == 0)
		{
			// ORIENT instruction
			os << "ORIENT " <<
			  GetDirectionString(memmap[address] ^ 0x00000020) << endl;
		}
		else
		{
			// CALL instruction
			unsigned calldest = (((memmap[address]&0x1f) + ((memmap[address+1]&0x1f)<<5))<<2);

			os << "CALL " << calldest << " ";
			
			os << "(" << revlabels[calldest + 4096] << ")";

			os << endl;

			address++;
		}
	}

	return address+1;
}

void OutputListing(char *f)
{
	ofstream os(f);

	if (os)
	{
		for(map<string,unsigned>::iterator i = labels.begin(); i != labels.end(); i++)
		{
			revlabels[(*i).second] = (*i).first;
		}

		unsigned address=0;

		while(address < addressend)
		{
			if (revlabels.find(address) != revlabels.end())
			{
				os << "     " << revlabels[address] << ":" << endl;
			}

			address = DisAssemble(os,address);
		}
	}
}

void OutputStimulus(char *f)
{
	unsigned iindex = 0;
	ofstream os(f);

	if (os)
	{
		char modname[] = "pc1.mc.mad.mem.memory-module-aa.x.xy";

		for(unsigned i = 0; i<32; i++)
		{
			if (i<26)
			{
				modname[29] = (char)('a'+i);
				modname[30] = '.';
				modname[31] = 'm';
				modname[32] = '.';
				modname[33] = 'i';
				modname[35] = (char)0;
				iindex = 34;
			}
			else
			{
				modname[29] = (char)('a'+i/26);
				modname[30] = (char)('a'+i%26);
				modname[31] = '.';
				modname[32] = 'm';
				modname[33] = '.';
				modname[34] = 'i';
				modname[36] = (char)0;
				iindex = 35;
			}

			for(unsigned j = 0; j<6; j++)
			{
				modname[iindex] = (char)('0'+j);

				os << modname << endl << "{" << endl;
				os << "  t=0" << endl;

				for(unsigned k = 0; k<256; k++)
				{
					if (memmap[i*256+(k+230+6*i+j)%256] & (1<<j))
					{
						os << "  0 1" << endl;
					}
					else
					{
						os << "  1 1" << endl;
					}
				}

				os << "  loop" << endl;
				os << "}" << endl;
			}
		}
	}
}

// Parameters: <listing> <stimulus>

int main(int argc, char *argv[])
{
	if (argc != 1 && argc != 2 && argc != 3)
	{
		cerr << "USAGE: constructorlanguage [<listing> [<stimulus>]]" << endl;
		return -1;
	}

	Assemble();	

	if (argc == 2 || argc == 3)
	{
		OutputListing(argv[1]);
	}

	if (argc == 3)
	{
		OutputStimulus(argv[2]);
	}

	// TODO - report any undefined references

	cout << "unsigned prog[] = {" << endl;

	for(unsigned a = 0; a<addressend; a++)
	{
		cout << (unsigned)(memmap[a]) << "," << endl;
	}

	cout << "};" << endl;

	cerr << addressend << endl;

	return 0;

}