#ifndef DTYPE_CPP
#define DTYPE_CPP

#include "signalpath.cpp"

// A D-type flip flop with Q output, data and load inputs
// Data input is assumed to be zero initialy
// load input is inverted, and assumed to be 1 initially (i.e. don't load)
// Both the nor gate at 2,0,1 and the one at 2,0,2 must have the load
// signal applied. This can be applied from north or south (hence the
// ld1 and ld2 terminals).
//
// Load input must be applied for 4 time units with data stable.
void DTypeFlipFlop(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nor {f e}}");
  bo.String("{del {b e}} {nor {f e}} {nor {w n}} {del {f e}}");
  bo.String("{del {w n}} {nor {w n}} {nor {w n}} {nor {w n}}");

  bo.SetOutputs(3,0,0,true);
  bo.SetOutputs(3,0,1,true);
  bo.SetOutputs(1,0,1,true);

  bo.Terminal("q",0,0,1,false,west);
  bo.Terminal("q2",0,0,1,false,back);
  bo.Terminal("d",3,0,0,true,back);
  bo.Terminal("ld1",2,0,2,true,north);
  bo.Terminal("ld2",2,0,2,true,south);
}

// A load path for use with the above flip flop - the load signal
// must be inverted. It gets applied to the right places -
// o1 matches up with ld1 above
// A load path can be sandwiched between two d types which require a common
// load signal
void DTypeLoadPath(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{del {f e}}");
  bo.String("{del {f e}}");
  bo.String("{del {f e}}");

  bo.SetOutputs(0,0,0,true);
  bo.SetOutputs(0,0,1,true);
  bo.SetOutputs(0,0,2,true);

  bo.Terminal("i",0,0,0,true,back);
  bo.Terminal("o1",0,0,2,false,south);
  bo.Terminal("o2",0,0,2,false,north);
}

// A series of n d-type flip flops, with a common load signal. n must be even
// You can specify whether you want the load signal to start at the north
// or the south end. The load signal is inverted
void DTypeSeries(CCompoundObject &co, unsigned n, bool loadnorth)
{
  std::string lpout,dtlpname,dtname;

  CBasicObject **dtypes = new CBasicObject *[n];
  CBasicObject **dtlps = new CBasicObject *[n/2];

  CBasicObject *loadpath;

  loadpath = new CBasicObject();

  if (loadnorth)
  {
    SignalPath(*loadpath,1,n*2-2,1,south,east,front,north,south,false,true);
  }
  else
  {
    SignalPath(*loadpath,1,n*2-2,1,north,east,front,south,north,false,true);
  }

  co.AddObject(loadpath,"loadpath",0,0,0);

  {for(unsigned i = 0; i<n; i++)
  {
    dtypes[i] = new CBasicObject();
    DTypeFlipFlop(*(dtypes[i]));

    if (i%2==1)
    {
      dtlps[i/2] = new CBasicObject();

      DTypeLoadPath(*(dtlps[i/2]));

      lpout = "oa";
      lpout[1] = 'a'+i/2;
      dtlpname = "lpa";
      dtlpname[2] = lpout[1];

      loadpath->Terminal(lpout.c_str(),0,2*i-2+(loadnorth?0:1),0,false,front);

      lpout = std::string("loadpath.") + lpout;

      co.AddObject(dtlps[i/2],dtlpname.c_str(),"i",lpout.c_str());

      dtname = "dta2";
      dtname[2] = 'a'+i/2;
     
      co.AddObject(dtypes[i-1],dtname.c_str(),"ld2",(dtlpname + std::string(".o2")).c_str());

      dtname[3] = '1';

      co.AddObject(dtypes[i],dtname.c_str(),"ld1",(dtlpname + std::string(".o1")).c_str());
    }
  }}

}

// A D-type flip flop with Q output, data and load inputs
// Data input is assumed to be zero initialy
// load input is inverted, and assumed to be 1 initially (i.e. don't load)
// Both the nor gate at 2,0,1 and the one at 2,0,2 must have the load
// signal applied. This can be applied from north or south (hence the
// ld1 and ld2 terminals).
//
// Load input must be applied for 4 time units with data stable.
void DTypeAdaptor(CBasicObject &bo)
{
  bo.Layer();
  bo.String("{nul {n e}} {nul {n e}} {nul {n e}} {nor {f e}}");
  bo.String("{nor {b e}} {nor {f e}} {nor {w n}} {del {f e}}");
  bo.String("{nor {b e}} {nor {f n}} {nor {f n}} {nor {w n}}");
  bo.String("{del {b e}} {del {f e}} {del {f e}} {nul {n e}}");

  bo.SetOutputs(3,0,0,true);
  bo.SetOutputs(3,0,1,true);
  bo.SetOutputs(1,0,1,true);
  bo.SetOutputs(0,0,2,true);

  bo.Terminal("q",0,0,1,false,back);
  bo.Terminal("d",3,0,0,true,back);
  bo.Terminal("ld1",2,0,2,true,north);
  bo.Terminal("ld2",2,0,2,true,south);
  bo.Terminal("from-setreset",0,0,3,true,front);
}

// A series of adaptors for d-type flip flops, with a common load signal.
// Modified from version below so that n can be odd (11/10/2008).
//
// You can specify whether you want the load signal to start at the north
// or the south end. The load signal is inverted
//
// This is designed to be the basis for a stack, in which the array of flip
// flops that get fitted to this adaptor can be shifted back and forth
// to store values in different locations in the stack.
void DTypeAdaptorSeries(CCompoundObject &co, unsigned n, bool loadnorth)
{
  std::string lpout,dtlpname,dtname;

  CBasicObject **adaptors = new CBasicObject *[n];
  CBasicObject **dtlps = new CBasicObject *[n-1];

  CBasicObject *loadpath;

  loadpath = new CBasicObject();

  if (loadnorth)
  {
    SignalPath(*loadpath,1,n*2-2,1,south,east,front,north,south,false,true);
  }
  else
  {
    SignalPath(*loadpath,1,n*2-2,1,north,east,front,south,north,false,true);
  }

  co.AddObject(loadpath,"loadpath",0,0,0);

  {for(unsigned i = 0; i<n; i++)
  {
    adaptors[i] = new CBasicObject();
    DTypeAdaptor(*(adaptors[i]));

    if (i != 0)
    {
      dtlps[i-1] = new CBasicObject();

      DTypeLoadPath(*(dtlps[i-1]));

      lpout = "oa";
      lpout[1] = 'a'+i;
      dtlpname = "lpa";
      dtlpname[2] = lpout[1];

      loadpath->Terminal(lpout.c_str(),0,2*i-2+(loadnorth?0:1),0,false,front);

      lpout = std::string("loadpath.") + lpout;

      co.AddObject(dtlps[i-1],dtlpname.c_str(),"i",lpout.c_str());

      dtname = "dta2";
      dtname[2] = 'a'+i;
     
      if (i==n-1)
      {
        co.AddObject(adaptors[i],dtname.c_str(),"ld2",(dtlpname + std::string(".o2")).c_str());
      }

      dtname[3] = '1';

      co.AddObject(adaptors[i-1],dtname.c_str(),"ld1",(dtlpname + std::string(".o1")).c_str());
    }
  }}

}


#if 0
// A series of adaptors for d-type flip flops, with a common load signal.
// n must be even
// You can specify whether you want the load signal to start at the north
// or the south end. The load signal is inverted
//
// This is designed to be the basis for a stack, in which the array of flip
// flops that get fitted to this adaptor can be shifted back and forth
// to store values in different locations in the stack.
void DTypeAdaptorSeries(CCompoundObject &co, unsigned n, bool loadnorth)
{
  std::string lpout,dtlpname,dtname;

  CBasicObject **adaptors = new CBasicObject *[n];
  CBasicObject **dtlps = new CBasicObject *[n/2];

  CBasicObject *loadpath;

  loadpath = new CBasicObject();

  if (loadnorth)
  {
    SignalPath(*loadpath,1,n*2-2,1,south,east,front,north,south,false,true);
  }
  else
  {
    SignalPath(*loadpath,1,n*2-2,1,north,east,front,south,north,false,true);
  }

  co.AddObject(loadpath,"loadpath",0,0,0);

  {for(unsigned i = 0; i<n; i++)
  {
    adaptors[i] = new CBasicObject();
    DTypeAdaptor(*(adaptors[i]));

    if (i%2==1)
    {
      dtlps[i/2] = new CBasicObject();

      DTypeLoadPath(*(dtlps[i/2]));

      lpout = "oa";
      lpout[1] = 'a'+i/2;
      dtlpname = "lpa";
      dtlpname[2] = lpout[1];

      loadpath->Terminal(lpout.c_str(),0,2*i-2+(loadnorth?0:1),0,false,front);

      lpout = std::string("loadpath.") + lpout;

      co.AddObject(dtlps[i/2],dtlpname.c_str(),"i",lpout.c_str());

      dtname = "dta2";
      dtname[2] = 'a'+i/2;
     
      co.AddObject(adaptors[i-1],dtname.c_str(),"ld2",(dtlpname + std::string(".o2")).c_str());

      dtname[3] = '1';

      co.AddObject(adaptors[i],dtname.c_str(),"ld1",(dtlpname + std::string(".o1")).c_str());
    }
  }}

}
#endif


#endif
