#ifndef OBJECTS_H
#define OBJECTS_H

#include <iostream>
#include <vector>
#include <string>
#include <map>

typedef enum {nul,delta,nor,push,slide,rotate,fuse,unfuse,keypress,signal,slideon,nopart} Part;

typedef enum {north,east,south,west,front,back,nodirection} Direction;

void SetConnect(bool c);

class CTerminal
{
  public:
    CTerminal(std::string n, int x, int y, int z, bool i, Direction f)
      : name(n), xpos(x), ypos(y), zpos(z), input(i), face(f) {}
  public:
    std::string name;
    int xpos,ypos,zpos;
    bool input;
    Direction face;
};

class CPartCell;

// TODO - don't forget the virtual destructors
class CObject
{
  public:
    virtual void Output(std::ostream &osg, std::ostream &ost, int x, int y, int z) = 0;
    virtual void Rotate(Direction a, int &xnew, int &ynew, int &znew, int x, int y, int z) = 0;
    virtual void Mirror(Direction a) = 0;
    virtual CTerminal *GetTerminal(const char *n, int &x, int &y, int &z) = 0;
    virtual CPartCell *GetPartCell(int x, int y, int z) = 0;
    virtual CObject *GetObject(const char *n) = 0;

    // Set the output values of the block, according to block type, orientation
    // and the specified value.
    void SetOutputs(int x, int y, int z, bool i);
    void WriteMem(char *part, char *t, unsigned offset, unsigned bits, unsigned *d, unsigned mask);

    virtual int XSize(void) = 0;
    virtual int YSize(void) = 0;
    virtual int ZSize(void) = 0;

    virtual int XMin(void) = 0;
    virtual int YMin(void) = 0;
    virtual int ZMin(void) = 0;

    virtual void ResolveConnections(CObject *topLevel, int x, int y, int z) = 0;

    virtual void Draw(std::ostream &os, int xo, int yo, int zo) = 0;
};

class CPartCell
{
  public:
    CPartCell(void)
      {part=nul; or1=north;or2=east;
       connect[0]=connect[1]=connect[2]=connect[3]=connect[4]=connect[5]=false;
       output[0]=output[1]=output[2]=output[3]=output[4]=output[5]=false;
      }
    CPartCell(Part p, Direction o1, Direction o2, bool *c, bool *o)
     : part(p), or1(o1), or2(o2)
       {
         connect[0] = c[0]; connect[1] = c[1]; connect[2] = c[2];
         connect[3] = c[3]; connect[4] = c[4]; connect[5] = c[5];

         output[0] = o[0]; output[1] = o[1]; output[2] = o[2];
         output[3] = o[3]; output[4] = o[4]; output[5] = o[5];
       }
    void Rotate(Direction a);
  public:
    Part part;
    Direction or1,or2;
    bool connect[6];
    bool output[6];
};

class CBasicObject : public CObject
{
  public:
    void Layer(void);
    bool String(const char *s);
    bool NulString(unsigned n);
    void Terminal(const char *n, int x, int y, int z, bool i, Direction f);

    // Rotate clockwise about the specified axis and return the x,y and z
    // coords (relative to this object) of the block that was previously
    // at 0,0,0
    void Rotate(Direction a, int &xnew, int &ynew, int &znew, int x, int y, int z);
    void Rotate(Direction a)
    {
      int xn,yn,zn;
      Rotate(a,xn,yn,zn,0,0,0);
    }

    void Mirror(Direction a);

    void SetAllOutputs(bool i);

    CTerminal *GetTerminal(const char *n,int &x, int &y, int &z);
    CPartCell *GetPartCell(int x, int y, int z);

    CObject *GetObject(const char *n);

    void WriteMem(char *part, char *t, unsigned offset, unsigned bits, unsigned *d, unsigned mask);

    void Output(std::ostream &osg, std::ostream &ost, int x, int y, int z);

    int XSize(void);
    int YSize(void);
    int ZSize(void);

    int XMin(void) {return 0;}
    int YMin(void) {return 0;}
    int ZMin(void) {return 0;}

    void ResolveConnections(CObject *topLevel, int x, int y, int z);

    void Draw(std::ostream &os, int xo, int yo, int zo);

  public:
    std::vector<std::vector<std::vector<CPartCell> > > cells;   // index as cells[y][z][x]
    std::vector<CTerminal> terminals;
};

class CObjectPosition
{
  public:
    CObjectPosition(CObject *o, std::string n, int x, int y, int z)
      : object(o), name(n), xpos(x), ypos(y), zpos(z) {}
  public:
    CObject *object;
    std::string name;
    int xpos,ypos,zpos;
};

class CCompoundObject : public CObject
{
  public:
    /* Add the object o to this compound object. Give it the name n (which
     * may be null if no name is required). Place it at coordinates x,y,z
     */
    void AddObject(CObject *o, const char *n, int x, int y, int z);

    /* Add the object o to this compound object. Give it the name n (which
     * may be null if no name is required). Align its terminal t (terminal
     * spec does not include n) with the terminal targt
     */
    bool AddObject(CObject *o, const char *n, const char *t, const char *targt);

    /* Rotate about axis a, passing through cell x,y,z */
    void Rotate(Direction a, int &xnew, int &ynew, int &znew, int x, int y, int z);
    void Rotate(Direction a, int x, int y, int z)
    {
      int xnew,ynew,znew;
      Rotate(a,xnew,ynew,znew,x,y,z);
    }
    void Rotate(Direction a)
    {
      Rotate(a,0,0,0);
    }

    void Mirror(Direction a);

    CObject *GetObject(const char *n);
    CTerminal *GetTerminal(const char *n, int &x, int &y, int &z);

    CPartCell *GetPartCell(int x, int y, int z);

    /* Set up a terminal alias, a is the alias name, t is the existing terminal
     */
    void Terminal(const char *a, const char *t);

    void Output(std::ostream &osg, std::ostream &ost, int x, int y, int z);

    int XSize(void);
    int YSize(void);
    int ZSize(void);

    int XMin(void);
    int YMin(void);
    int ZMin(void);

    void ResolveConnections(CObject *topLevel, int x, int y, int z);

    void Draw(std::ostream &os, int xo, int yo, int zo);
  private:
    std::map<std::string,std::string> terminalalias;
    std::vector<CObjectPosition> objects;
    
};


#endif
