#ifndef UNIVERSE_H
#define UNIVERSE_H

#include <map>
#include <list>
#include <fstream>
#include <string>
#include <vector>

#include "Block.h"

#define ATTENTIONNEXT_MASK 1
#define DOUPDATE_MASK 2
#define DOMOVE_MASK 4
#define ATTENTIONNOW_MASK 8

#define CELL_X 2
#define CELL_Y 2
#define CELL_Z 3

#define PROC_N_MASK 1
#define PROC_S_MASK 2
#define PROC_E_MASK 4
#define PROC_W_MASK 8
#define PROC_F_MASK 16
#define PROC_B_MASK 32

#define PROC_ALL_MASK 63

// regions are used to identify which parts belong to which subsystem. Bookkeeping only
class CRegion
{
public:
	CRegion(char *c, CRegion *p) :  parent(p), visible(true), visiblecheck(true)
		{
		  if (p)
		  {
			  id = p->id + std::string(".") + std::string(c);
		  }
		  else
		  {
			  id = std::string(c);
		  }
		}

	CRegion *GetParent(void) {return parent;}
	std::string GetID(void) {return id;}
	bool IsVisible(void) {return visible;}
	bool IsVisibleCheck(void) {return visiblecheck;}
	void SetVisible(bool v, bool c) {visible = v; visiblecheck = c;}
private:
	std::string id;
	CRegion *parent;
	bool visible;
	bool visiblecheck;
};


class CUniverse;

class CUniverseCell
{
public:
	CUniverseCell(int x, int y, int z, CUniverse *p);
	~CUniverseCell();

    void AddBlock(CBlock *b);
	CBlock *GetBlock(int x, int y, int z);

	bool GetSignalValue(int x, int y, int z, Direction d);

	unsigned Process(void);
	bool DoMove(void);
	bool Update(void);

	unsigned Move(int x, int y, int z, CMove *m);

	void Save(std::ofstream &os);

	void ResolveConnections(void);
	void SetNeighbours(void);

	unsigned Draw(void);
	void WriteEPS(std::ofstream &os);

	bool Unused(void);

	void FindOrigin(std::vector<triple<int, int, int> > &candidates);
	void CentreOfMass(double &x, double &y, double &z, unsigned &n);
public:

	int xpos;
	int ypos;
	int zpos;
	CBlock *blocks[CELL_X][CELL_Y][CELL_Z];
	CUniverse *parent;

	unsigned listbits;
};

#if 1
#define HASH_ARRAY_SIZE 99991


class SimpleUCList
{
public:
	SimpleUCList(void) : uc(NULL), next(NULL) {}

	CUniverseCell *uc;
	SimpleUCList *next;
};

class CellsHolder
{
public:

    unsigned HASH(int x, int y,int z)
	{
		return ( ( (unsigned) ( ((x+30)<<5) ^ ((y+30)<<10) ^ (z+30)) ) ) % HASH_ARRAY_SIZE;
	}

	CellsHolder(void)
	{
		for(unsigned i = 0; i<HASH_ARRAY_SIZE; i++)
		{
			hash[i] = NULL;
		}
	}

	~CellsHolder(void)
	{
		SimpleUCList *p,*plast;

		for(unsigned i = 0; i<HASH_ARRAY_SIZE; i++)
		{
			p = hash[i];
			
			while(p)
			{
				delete p->uc;

				plast = p;
				p = p->next;

				delete plast;
			}
		}
	}

	CUniverseCell *Find(int x, int y, int z)
	{
		SimpleUCList *p = hash[HASH(x,y,z)];

		while(p)
		{
			if (p->uc->xpos == x*CELL_X && p->uc->ypos == y*CELL_Y && p->uc->zpos == z*CELL_Z)
			{
				return p->uc;
			}

			p = p->next;
		}

		return NULL;
	}

	void Add(int x, int y, int z, CUniverseCell *c)
	{
		SimpleUCList *p = hash[HASH(x,y,z)];

		if (!p)
		{
			p = hash[HASH(x,y,z)] = new SimpleUCList;
			p->uc = c;
			return;
		}

		while(p)
		{
			if (p->uc->xpos == x*CELL_X && p->uc->ypos == y*CELL_Y && p->uc->zpos == z*CELL_Z)
			{
				// should never reach here
				p->uc = c;
				return;
			}

			if (!p->next)
			{
				p->next = new SimpleUCList;
				p->next->uc = c;
				return;
			}

			p = p->next;
		}
	}

	void Save(std::ofstream &os)
	{
		SimpleUCList *p;

		for(unsigned i = 0; i<HASH_ARRAY_SIZE; i++)
		{
			p = hash[i];
			
			while(p)
			{
				p->uc->Save(os);

				p = p->next;
			}
		}
	}

	void ResolveConnections(void)
	{
		SimpleUCList *p;

		for(unsigned i = 0; i<HASH_ARRAY_SIZE; i++)
		{
			p = hash[i];
			
			while(p)
			{
				p->uc->ResolveConnections();

				p = p->next;
			}
		}
	}

	void SetNeighbours(void)
	{
		SimpleUCList *p;

		for(unsigned i = 0; i<HASH_ARRAY_SIZE; i++)
		{
			p = hash[i];
			
			while(p)
			{
				p->uc->SetNeighbours();

				p = p->next;
			}
		}
	}

	unsigned Draw(void)
	{
		unsigned total = 0;

		SimpleUCList *p;

		for(unsigned i = 0; i<HASH_ARRAY_SIZE; i++)
		{
			p = hash[i];
			
			while(p)
			{
				total += p->uc->Draw();

				p = p->next;
			}
		}

		return total;
	}


	void WriteEPS(std::ofstream &os)
	{
		SimpleUCList *p;

		for(unsigned i = 0; i<HASH_ARRAY_SIZE; i++)
		{
			p = hash[i];
			
			while(p)
			{
				p->uc->WriteEPS(os);

				p = p->next;
			}
		}
	}

	void Clean(void)
	{
		SimpleUCList *p;
		SimpleUCList *last;

		for(unsigned i = 0; i<HASH_ARRAY_SIZE; i++)
		{
			p = hash[i];
			
			while(p)
			{
				// only delete if it is not going to be in the attentionnow list next cycle.
				if ( (!(p->uc->listbits & ATTENTIONNOW_MASK)) && p->uc->Unused() )
				{
					delete p->uc;
						
					if (p!=hash[i])
					{
						last->next = p->next;
						delete p;
						p = last->next;
					}
					else
					{
						hash[i] = p->next;
						delete p;
						p = hash[i];
					}
				}
				else
				{
					last = p;
					p = p->next;
				}
			}
		}
	}

	void FindOrigin(std::vector<triple<int,int,int> > &candidates)
	{
		SimpleUCList *p;

		for(unsigned i = 0; i<HASH_ARRAY_SIZE; i++)
		{
			p = hash[i];
			
			while(p)
			{
				p->uc->FindOrigin(candidates);

				p = p->next;
			}
		}
	}

	void CentreOfMass(double &x, double &y, double &z)
	{
		SimpleUCList *p;
		x = 0; y = 0; z = 0;
		unsigned n = 0;

		for(unsigned i = 0; i<HASH_ARRAY_SIZE; i++)
		{
			p = hash[i];
			
			while(p)
			{
				p->uc->CentreOfMass(x,y,z,n);

				p = p->next;
			}
		}

		x = x / ((float)n);
		y = y / ((float)n);
		z = z / ((float)n);
	}

private:
	SimpleUCList *hash[HASH_ARRAY_SIZE];
};

#else // old version of CellsHolder that uses a map
class CellsHolder
{
public:
	~CellsHolder(void)
	{
		for(std::map< triple<int,int,int> , CUniverseCell *>::iterator i = cells.begin(); i!=cells.end(); i++)
		{
			delete (*i).second;
		}
	}

	CUniverseCell *Find(int x, int y, int z)
	{
		std::map< triple<int,int,int> , CUniverseCell *>::iterator it;

		it = cells.find( triple<int,int,int>(x,y,z));

		if (it == cells.end())
		{
			return NULL;
		}
		else
		{
			return it->second;
		}
	}

	void Add(int x, int y, int z, CUniverseCell *c)
	{
		cells[ triple<int,int,int>(x,y,z) ] = c;
	}

	void Save(std::ofstream &os)
	{
		for(std::map< triple<int,int,int>, CUniverseCell *>::iterator i = cells.begin(); i!=cells.end(); i++)
		{
			(*i).second->Save(os);
		}
	}

	void ResolveConnections(void)
	{
		for(std::map< triple<int,int,int>, CUniverseCell *>::iterator i = cells.begin(); i!=cells.end(); i++)
		{
			(*i).second->ResolveConnections();
		}
	}

	unsigned Draw(void)
	{
		unsigned total = 0;

		for(std::map< triple<int,int,int>, CUniverseCell *>::iterator i = cells.begin(); i!=cells.end(); i++)
		{
			total += (*i).second->Draw();
		}

		return total;
	}


	void WriteEPS(std::ofstream &os)
	{
		for(std::map< triple<int,int,int>, CUniverseCell *>::iterator i = cells.begin(); i!=cells.end(); i++)
		{
			(*i).second->WriteEPS(os);
		}
	}

	// remove any unused universecells
	void Clean(void)
	{
		for(std::map< triple<int,int,int>, CUniverseCell *>::iterator i = cells.begin(); i!=cells.end();)
		{
			// only delete if it is not going to be in the attentionnow list next cycle.
			if ( (!((*i).second->listbits & ATTENTIONNOW_MASK)) && (*i).second->Unused() )
			{
				delete (*i).second;
				i = cells.erase(i);
			}
			else
			{
				i++;
			}
		}
	}

private:
	std::map< triple<int,int,int> , CUniverseCell *> cells;
};
#endif

class CUniverse
{
public:
	CUniverse(void);

	~CUniverse();

    void AddBlock(CBlock *b);
	CBlock *GetBlock(int x, int y, int z);
	CBlock *GetBlockNoAdd(int x, int y, int z);

	bool GetSignalValue(int x, int y, int z, Direction d);
	void ApplyStimulus(int x, int y, int z, Direction f);

	void Process(void);
	void DoMove(void);
	void Update(void);
	void Clean(void);

	unsigned Move(int x, int y, int z, CMove *m);
	void MakeMove(int x, int y, int z, Direction d);
	void Rotate(int x, int y, int z, Direction d);
	void Connect(int x, int y, int z, Direction d, bool c);

	void Draw(void);
	void WriteEPS(std::ofstream &os);

	void Load(std::ifstream &is);
	void Save(std::ofstream &os);

	void ResolveConnections(void);
	void SetNeighbours(void);


	void MarkCells(int x, int y, int z, unsigned procmask, bool nextOrNow = true);

	void InsertIntoDoUpdate(CUniverseCell *c);
	void InsertIntoDoMove(CUniverseCell *c);
	void InsertIntoAttentionNow(CUniverseCell *c);
	void InsertIntoAttentionNext(CUniverseCell *c);

	std::list<CRegion *> &GetRegions(void) {return regions;}
	std::string GetRegionName(int x, int y, int z);

	void WriteToMemory(int x, int y, int z, int offset, std::vector<unsigned> &data);
	void WriteWordToMemoryBlock(int x, int y, int z, unsigned pos, unsigned word);

	// Find a location to use as an origin - use the x,y,z coords of the front-south
	// most rotate part that 
	void FindOrigin(int &x, int &y, int &z);
	void CentreOfMass(double &x, double &y, double &z);
public:
	bool press;

	CellsHolder cells;
	std::list<CMove *> moves;

	std::list<CUniverseCell *> l1,l2;

	std::list<CUniverseCell *> *attentionNow, *attentionNext;
	std::list<CUniverseCell *> doMove, doUpdate;

	unsigned iters;
	unsigned totalDrawn;

	// below is for diagnostics only;
	int lastmovex,lastmovey,lastmovez;
	int activeCount;

	std::list<CRegion *> regions;

	bool moveView;
	Direction moveViewDir;
};

#endif